/*****************************/
/**       ------------      **/
/**        camera.cpp       **/
/**       ------------      **/
/**  Inicializacin y ren-  **/
/**  derizacin de modelos  **/
/*****************************/

/*** Estructura de datos: CAMERA ***/
typedef struct camera
{
  D3DXVECTOR3 pos;
  D3DXVECTOR3 right;
  D3DXVECTOR3 up;
  D3DXVECTOR3 look;
} CAMERA;
/*_______*/


/*** Funcin: Movimiento lateral ***/
void Strafe( CAMERA* cam, float units, bool yLock )
{
  if( yLock )
    cam->pos += D3DXVECTOR3( cam->right.x, 0.0f, cam->right.z ) * units;
  else
    cam->pos += cam->right * units;
}

/*** Funcin: Movimiento altitudinal ***/
void Fly( CAMERA* cam, float units )
{
  cam->pos += cam->up * units;
}

/*** Funcin: Movimiento frontal ***/
void Walk( CAMERA* cam, float units, bool yLock )
{
  if( yLock )
    cam->pos += D3DXVECTOR3( cam->look.x, 0.0f, cam->look.z ) * units;
  else
    cam->pos += cam->look * units;
}

/*** Funcin: Rotacin en el eje "derecha" ***/
void Pitch( CAMERA* cam, float units )
{
  D3DXMATRIX T;
  D3DXMatrixRotationAxis( &T, &cam->right, units );
  
  D3DXVec3TransformCoord( &cam->up, &cam->up, &T );
  D3DXVec3TransformCoord( &cam->look, &cam->look, &T );

  D3DXVec3Normalize( &cam->up, &cam->up );
  D3DXVec3Normalize( &cam->look, &cam->look );
}

/*** Funcin: Rotacin en el eje "arriba" ***/
void Yaw( CAMERA* cam, float units, bool upYaw )
{
  D3DXMATRIX T;
  if( upYaw )
    D3DXMatrixRotationY( &T, units );
  else
    D3DXMatrixRotationAxis( &T, &cam->up, units );
  
  D3DXVec3TransformCoord( &cam->right, &cam->right, &T );
  D3DXVec3TransformCoord( &cam->look, &cam->look, &T );

  D3DXVec3Normalize( &cam->right, &cam->right );
  D3DXVec3Normalize( &cam->look, &cam->look );
}

/*** Funcin: Rotacin en el eje "direccin" ***/
void Roll( CAMERA* cam, float units )
{
  D3DXMATRIX T;
  D3DXMatrixRotationAxis( &T, &cam->look, units );
  
  D3DXVec3TransformCoord( &cam->right, &cam->right, &T );
  D3DXVec3TransformCoord( &cam->up, &cam->up, &T );
  
  D3DXVec3Normalize( &cam->right, &cam->right );
  D3DXVec3Normalize( &cam->up, &cam->up );
}
